from .borders import Borders, BorderDfn
from .card import Card
from .fill import Fill
from .flex import Flex
from .hr import HR
from .line import Line
from .pretty import pretty
from .printer import Printer
from .table import Table, CellDfn
from .text import Text
from . import context
print = Printer()



from .table import Table


def pretty(val, _depth=0, **kwargs):
    if isinstance(val, tuple): return _pretty_list(val, '()', _depth, **kwargs) if val else 'tuple()'
    if isinstance(val, list):  return _pretty_list(val, '[]', _depth, **kwargs) if val else '[]'
    if hasattr(val, 'items'): return _pretty_dict(val, _depth, **kwargs) if val else '{}'
    if hasattr(val, 'pretty'): return val.pretty(**kwargs)
    return val


def _pretty_list(val, paren, _depth, tmpl='', **kwargs):
    t = Table(0.0, 0, tmpl=tmpl, **kwargs)
    t.cell('C0', style=str(_depth%3 + 1), just='>')
    for i,v in enumerate(val):
        t(f'\bdem {paren[0]}', i,f'\bdem {paren[1]} ', '\t', pretty(v, _depth+1), '\t')
    return t


def _pretty_dict(val, _depth, tmpl='', **kwargs):
    t = Table(0.0, 0, tmpl=tmpl, **kwargs)
    t.cell('C0', style=str(_depth%3 + 1), just='>')
    for k,v in val.items():
        t(k,' ', '\t', pretty(v, _depth+1),'\t')
    return t



import pytest
from print_ext.card import Card
from print_ext.line import SMark as SM
from print_ext.flex import Flex
from .testutil import flat, printer, styled


def test_card_hello():
    c = Card('hello\t', 'to the great big\vworld')
    assert(flat(c) == [
        '┌┤ hello ├─────────┐',
        '│ to the great big │',
        '│ world            │',
        '└──────────────────┘',])


def test_card_multi_title():
    assert(flat(Card('a\vb\vc\tinner body\t')) == [
        ' │ a │        ',
        ' │ b │        ',
        '┌┤ c ├───────┐',
        '│ inner body │',
        '└────────────┘',   
    ])


def test_card_empty():
    assert(flat(Card()) == ['┌──┐', '└──┘'])


def test_card_notitle():
    assert(flat(Card('\tHello to the\vworld')) == [
        '┌──────────────┐', 
        '│ Hello to the │', 
        '│ world        │', 
        '└──────────────┘'])



def test_card_clone():
    c = Card('title\tbody')
    f = Flex(c, '\t', c)
    assert(flat(f) == ['┌┤ title ├┐┌┤ title ├┐', '│ body    ││ body    │', '└─────────┘└─────────┘'])



def test_border_hilite_bug():
    ''' Show a portion of the table while building.  The col widths may change, so use fixed col widths
    '''
    o,p = printer(color=True)
    p.card('\berr Error:\t', 'Some problem', border_style='y')
    c = Card('\berr Error:\t', 'Some problem')
    
    print(o.getvalue())
    print(repr(o.getvalue()))
    assert(styled(c) == [
        ('┌┤ Error: ├────┐', [SM('dem',0,3), SM('err',3,9), SM('dem',9,16)]),
        ('│ Some problem │', [SM('dem',0,2), SM('dem',14,16)]),
        ('└──────────────┘', [SM('dem',0,16)]),
    ])


@pytest.mark.xfail(reason="implement")
def test_card_obey_max_width():
    assert(False)

    


from .context import Context, IntCVar, ObjectAttr


INFINITY = 1000000000000

Context.define(IntCVar('width_nom', 'ww'))
Context.define(IntCVar('width_min', 'wm'))
Context.define(IntCVar('width_max', 'wx'))
Context.define(IntCVar('height_nom', 'hh'))
Context.define(IntCVar('height_min', 'hm'))
Context.define(IntCVar('height_max', 'hx'))



class Widget(Context, width_nom=ObjectAttr('width', None), height_nom=ObjectAttr('height', None)):
    ''' The obligatory widget; Generic ancestor to all things flattenable.

    This is an abstract base class for all things that draw to a rectangle.
    '''

    def __init__(self, *args ,**kwargs):
        super().__init__(*args, **kwargs)


    @property
    def width(self):
        return self.calc_width()


    @property
    def height(self):
        return self.calc_height()


    def calc_width(self, klass=None):
        flat = list(klass.flatten(self) if klass else self.flatten())
        return flat[0].width if flat else 0


    def calc_height(self, klass=None):
        flat = list(klass.flatten(self) if klass else self.flatten())
        return len(flat)


    def changed_size(self):
        self.__width = None
        #self._clear__width()
        #self._clear__height()
    

    def __repr__(self):
        return f"<{self.__class__.__name__}>"



from wcwidth import wcswidth
from collections import namedtuple
from functools import reduce

SingleSpan = namedtuple('SingleSpan', ['text', 'width', 'charsize'])


class Span():
    ''' A Span represents a line of characters as a list of SingleSpan objects.

    It is immutable, as are SingleSpans.
    
    A singleSpan object helps by separating homogenious chunks of characters.
    A Span does not contain line breaks.  If a newline character is added
    to a Span then it will be renderd in escaped form '\'+'n'
    '''
    __slots__ = ('length', 'width', 'spans')

    def __init__(self, *args):
        self.spans = []
        self.length = 0
        self.width = 0
        for arg in args:
            if isinstance(arg, Span):
                self.spans += arg.spans
                self.width += arg.width
                self.length += arg.length
                self._merge_two(len(self.spans) - len(arg.spans) - 1)
                continue
            for span in [arg] if isinstance(arg, SingleSpan) else _SingleSpan_from_str(str(arg)):
                self.width += span.width
                self.length += len(span.text)
                self.spans.append(span)
                self._merge_two(len(self.spans) - 2)
        

    def trim(self, width):
        ''' Return a Span that fits into `width`.

        If `width` >=0 then keep the left.
        If `width` < 0 then keep the right.
        '''
        if not self: return Span()
        i, idx = self._find_char_at(*((self.width+width, 1) if width < 0 else (width,0)))
        if idx != 0:
            span = self.spans[i]
            i += 1 if idx>=len(span.text) else 1 if self._split_at(span, i, idx) == None else 0
        return Span(*(self.spans[i:] if width < 0 else self.spans[:i]))


    def ctx_parent(self, p):
        return self


    def _merge_two(self, idx):
        if idx < 0 or idx + 1 >= len(self.spans): return
        pA = self.spans[idx]
        pB = self.spans[idx+1]
        if pA.charsize == pB.charsize:
            self.spans[idx:idx+2] = [SingleSpan(text=pA.text+pB.text, width=pA.width + pB.width, charsize=pA.charsize)]
            return True
        

    def _find_in_span(self, span, i, w, dir):
        if span.charsize != 0: return span, i, w//span.charsize + (w%span.charsize and dir)
        if (aW:=self._split_at(span, i, len(span.text) // 2)) != None:
            i -= 1
            w += self.spans[i].width - aW
        spanA = self.spans[i]
        return self._find_in_span(spanA, i, w, dir) if w < spanA.width else self._find_in_span(self.spans[i+1], i+1, w-spanA.width, dir)
    

    def _find_char_at(self, width, dir=0):
        assert(self.spans)
        if width < 0: return 0,0
        for i, span in enumerate(self.spans):
            if span.width > width: break
            width -= span.width
        else:
            return i+1, 0
        span, i, idx = self._find_in_span(span, i, width, dir)
        return (i+1, 0) if idx >= len(span.text) else (i, idx)
        

    def _split_at(self, span, i, idx):
        spanA = _new_span(text=span.text[:idx])
        assert(spanA.text)
        spanB = _new_span(text=span.text[idx:])
        assert(spanB.text)
        self.spans[i:i+1] = [spanA, spanB]
        self._merge_two(i+1)
        return spanA.width if self._merge_two(i-1) else None
         

    def __bool__(self):
        return bool(self.spans)


    def __len__(self):
        return self.length


    def __str__(self):
        return reduce(lambda a,s: a + _str_span(s), self.spans, '')


    def __add__(self, other):
        return Span(self, other)


    def __radd__(self, other):
        return Span(other, self)


    def __repr__(self):
        spans = [repr(s.text) for s in self.spans]
        return f"Span({', '.join(spans)})"




def _SingleSpan_from_str(t):
    w = wcswidth(t)
    if w < 0:
        if len(t) == 1:
            w = len(_chr(t[0]))
            yield SingleSpan(text=t, width=w, charsize=w)
        else:
            m = len(t)//2
            yield from _SingleSpan_from_str(t[:m])
            yield from _SingleSpan_from_str(t[m:])
    else:
        if t: yield _new_span(text=t, width=w)



def _new_span(text, width=None):
    lt = len(text)
    w = wcswidth(text) if width==None else width
    assert(w >= 0)
    assert(text)
    return SingleSpan(text=text, width=w, charsize=(1 if w==lt else 2 if w==2*lt else 0))



def _str_span(span):
    return span.text if span.charsize <= 2 else ''.join(map(_chr, span.text))



def _chr(c):
    o = ord(c)
    if o == 9: return '    '
    if o < 256: return f'\\x{o:02x}'
    if o < 65536: return f'\\u{o:04x}'
    return repr(chr(o))[1:-1]





from print_ext.hr import HR
from print_ext.line import SMark as SM
from print_ext.flex import Flex
from print_ext.text import Text
from .printer_test import _printer
from .testutil import flat, printer, styled

def test_hr_hello_pretty():
    #o, p = _printer(width=10, ascii=True)
    hr = HR('\b2 hi', border_style='1', ascii=True)
    assert(styled(hr, w=10)[0] == ('-[ hi ]---', [SM('1',0,3), SM('2',3,5), SM('1',5,10)]))


def test_hr_blank():
    assert(flat(HR(width_max=10, ascii=True)) == ['----------'])
    assert(flat(HR(border=('t:abcde'))) == ['acd'])
    assert(HR().width == 3)
    assert(HR(width_max=5).width == 5)
    assert(flat(HR('','',width_max=4,ascii=True)) == ['----'])


def test_hr_just_h():
    o,p = _printer(width=10, ascii=True)
    p.hr('hi', just='<')
    print(o.getvalue())
    assert(o.getvalue() == '-[ hi ]---\n')
    h = HR('hi', just='|', ascii=True, width_max=10, style='y')
    assert([f.styled() for f in h.flatten()][0] == ('--[ hi ]--', [SM('y',0,10)]))
    assert([f.styled() for f in h.flatten(w=9)][0] == ('-[ hi ]--', [SM('y',0,9)]))
    h['just'] = ':'
    assert([f.styled() for f in h.flatten(w=9)][0] == ('--[ hi ]-', [SM('y',0,9)]))
    print('\n\n\n\n')
    h['just'] = '>'
    assert([f.styled() for f in h.flatten()][0] == ('---[ hi ]-', [SM('y',0,10)]))



def test_hr_just_v():
    h = HR('a\vbbb\vc\v', ascii=True, width_max=11)
    assert([f.styled()[0] for f in h.flatten()] == [' [ a   ]   ', ' [ bbb ]   ','-[ c   ]---',' [     ]   '])
    h['just'] = '_'
    assert([f.styled()[0] for f in h.flatten()] == [' [ a   ]   ', ' [ bbb ]   ',' [ c   ]   ','-[     ]---'])
    h['just'] = '|^'
    assert([f.styled()[0] for f in h.flatten()] == ['--[  a  ]--', '  [ bbb ]  ','  [  c  ]  ','  [     ]  '])
    h['just'] = '|-'
    h['ascii'] = False
    assert([f.styled()[0] for f in h.flatten()] == ['  │  a  │  ', '──┤ bbb ├──','  │  c  │  ','  │     │  '])
    h['just'] = '>'
    assert([f.styled()[0] for f in h.flatten()] == ['   │   a │ ', '   │ bbb │ ','───┤   c ├─','   │     │ '])



def test_hr_small():
    h = HR('hello world', ascii=True, wrap=False)
    assert([f.styled()[0] for f in h.flatten()] == ['-[ hello world ]-'])
    assert([f.styled()[0] for f in h.flatten(w=11)] == ['-[ he~ld ]-'])
    assert([f.styled()[0] for f in h.flatten(w=7)] == ['-[ ~ ]-'])
    assert([f.styled()[0] for f in h.flatten(w=6)] == ['hel~ld'])
    h['wrap']=True
    assert([f.styled()[0] for f in h.flatten(w=11)] == [' [ hello ] ', '-[ \\  wo ]-', ' [ \\ rld ] '])



def test_hr_clone():
    h = HR('bob', ascii=False)
    print(h['width_max'])
    assert([str(x) for x in h.flatten()] == ['─┤ bob ├─'])
    assert([str(x) for x in Text(h,h).flatten()] == ['─┤ bob ├──┤ bob ├─'])


def test_hr_infinity():
    o,p = printer(ascii=True)
    p.hr()
    assert(o.getvalue() == '---\n')
